import datetime
from Products.CMFCore.utils import getToolByName
from tecnoteca.plonecmdbuild.ws.SignatureHandler import SignatureHandler
from tecnoteca.plonecmdbuild.ws.WebservicesImplService_services import *
from tecnoteca.plonecmdbuild.config import PROJECTNAME
from tecnoteca.plonecmdbuild.ws.WSProperties import WSProperties
from tecnoteca.plonecmdbuild import plonecmdbuildMessageFactory as _
from tecnoteca.plonecmdbuild.browser.logger import logger
from zope.interface import implements
from elementtree.ElementTree import Element, SubElement
from tecnoteca.plonecmdbuild.ws.WSConst import WSConst as const
from ZTUtils.Zope import make_query
from zExceptions import Redirect
from pprint import pformat

   
class CMDBuildWebservice: 
    
    def _wsMethod(wsMethod, *args, **kwargs):
        """ decorator to handle ws methods """
        def wsmWrapper(self, *args, **kwargs):
            try:
                name = wsMethod.__name__
                logger.debug('User %s: calling "%s" with args %s and kwargs %s' % (self.username, name, pformat(args), pformat(kwargs)))
                response = wsMethod(self, *args, **kwargs)
                logger.debug('User %s: response %s' % (self.username, pformat(response)))
                return response
            except Exception, err:
                logger.exception(err)
                url = self.context.absolute_url()
                pt = "cmdbuildError"
                href = '%s/%s?%s' % (url, pt, make_query({const.ERROR:err}))
                raise Redirect, (href)
        return wsmWrapper
    
    
    def __init__(self, context):
        self.context = context
        self.request = self.context.REQUEST
        self.session = self.request.SESSION
                     
        # read current username
        self.username = self.getPloneUsername()
        
        # read properties
        prop = WSProperties()
        
        # init ws
        sigHandler = SignatureHandler(prop, self.username)
        wsLocator = WebservicesImplServiceLocator(prop.getWSUrl())        
                    
        # setup conn
        self.dbfetchSrv = wsLocator.getWebservicesPortType()
        self.dbfetchSrv.binding.sig_handler = sigHandler
        
        # session
        self.loadMenu()
        self.loadFlowStatuses()
        

    def loadFlowStatuses(self):        
        flowStatuses = self.session.get(const.SESSION_FLOW_STATUSES)
        if(flowStatuses == None):
            flowStatuses = self.getLookupList('FlowStatus')
            self.session.set(const.SESSION_FLOW_STATUSES, flowStatuses)
            logger.debug("User %s: session var '%s' set to %s" % (self.username, const.SESSION_FLOW_STATUSES, flowStatuses))
        return flowStatuses
            
            
    def loadMenu(self):
        cmdbuildMenu = self.session.get(const.SESSION_MENU)
        if(cmdbuildMenu == None):
            cmdbuildMenu = self.getMenuSchema()
            self.session.set(const.SESSION_MENU, cmdbuildMenu)
            logger.debug("User %s: session var '%s' set to %s" % (self.username, const.SESSION_MENU, cmdbuildMenu))
        return cmdbuildMenu
            
    
    def loadCMDBuildUser(self, metadataValue):
        cmdbuildUser = self.session.get(const.SESSION_CMDBUILD_USER)
        if(cmdbuildUser == None):
            cmdbuildUser = self.getCMDBuildUser(metadataValue)
            self.session.set(const.SESSION_CMDBUILD_USER, cmdbuildUser)
            logger.debug("User %s: session var '%s' set to %s" % (self.username, const.SESSION_CMDBUILD_USER, cmdbuildUser))
        return cmdbuildUser
    
    
    def setSearchDict(self, attributes):
        searchDict = dict()
        for attribute in attributes:
            if self.request.has_key(attribute['name'] + '_searchname'):
                searchDict[attribute['name']] = self.request.get(attribute['name'] + '_comparator') + '|' + self.request.get(attribute['name'] + '_searchname')        
        self.session.set(const.SESSION_SEARCHDIC, searchDict)
        logger.debug("User %s: session var '%s' set to %s" % (self.username, const.SESSION_SEARCHDIC, searchDict))
        return searchDict
                
    
    def clearSessionVars(self):
        """ clear plonecmdbuild session vars """
        sessionVars = [const.SESSION_CMDBUILD_USER, const.SESSION_SEARCHDIC, const.SESSION_SEARCHDIC, const.SESSION_FLOW_STATUSES, const.SESSION_MENU]
        for var in sessionVars:
            self.session.set(var, None)
        logger.debug("User %s: removed session vars: %s" % (self.username, sessionVars))
            
    
    def getPloneUsername(self):
        portal_membership = getToolByName(self.context, 'portal_membership')
        return portal_membership.getAuthenticatedMember().getId()
            
   
        
        
    """ *********************************************** ws methods ************************************************** """        

    @_wsMethod
    def getClassDetails(self,className):
                
        def searchNode(nodes):
            foundNode = None
            if nodes:                
                for node in nodes:
                    if node[const.CLASSNAME] == className:
                        foundNode = node
                    elif node[const.MENUTYPE] == const.FOLDERMENUTYPE:
                        foundNode = searchNode(node[const.NODES])
                        
                    if foundNode:
                        break
                    
            return foundNode
        
        # init
        entry = dict()
        entry[const.CLASSNAME] = None
        entry[const.DESCRIPTION] = None
        entry[const.META_PRIVILEGES] = None
        entry[const.ISPROCESS] = None
        entry[const.FLOWSTATUS] = None
        
        if className:
            # get node
            request = self.context.REQUEST
            wsMenu = self.session.get(const.SESSION_MENU)
            node = searchNode(wsMenu)
            
            if node:
                
                #classname
                entry[const.CLASSNAME] = node[const.CLASSNAME]
                
                #description
                entry[const.DESCRIPTION]=node[const.DESCRIPTION]
                
                # privileges
                privilege=""
                
                if const.META_PRIVILEGES in node:
                    privilege=node[const.META_PRIVILEGES]
                entry[const.META_PRIVILEGES] = privilege
                
                # is process
                isProcess = False
                if node[const.MENUTYPE] == const.PROCESSMENUTYPE:
                    isProcess = True
                entry[const.ISPROCESS]=isProcess
                
                #flowstatus
                if self.request.get(const.FLOWSTATUS)!=None:
                    entry[const.FLOWSTATUS] = self.request.get(const.FLOWSTATUS)
                else:                    
                    if isProcess:
                        openRunningId = -1
                        openRunningStr = "open.running"
                        flowStatuses = self.session.get(const.SESSION_FLOW_STATUSES)
                        for row in flowStatuses:
                            if row[const.CODE] == openRunningStr:
                                openRunningId = row[const.ID]
                        if openRunningId == -1:
                            raise Exception(_(u"No id for status:") + openRunningStr)
                        entry[const.FLOWSTATUS] = openRunningId
                    else:
                        entry[const.FLOWSTATUS] = 0
                
                # check privileges
                if className=="None":
                    raise Exception(_(u"Insufficient privileges"))
            
            else:
                # not found
                raise Exception(_(u"Class not found in menu:") + className)
                
        return entry     

    
    @_wsMethod
    def getMenuSchema(self):
        request = getMenuSchema()
        response = self.dbfetchSrv.getMenuSchema(request)        
    
        def getItems(items):
            if not items:
                return ''
            
            ret = []
            if items:
                for c in items:
                    cl = dict()
                    cl[const.DESCRIPTION] = c._description
                    cl[const.CLASSNAME] = c._classname
                    cl[const.ID] = c._id
                    cl[const.PARENT_ID] = c._parentId
                    cl[const.POSITION] = c._position
                    cl[const.MENUTYPE] = c._menuType
                    for meta in c._metadata:
                        cl[meta._key] = meta._value
                    cl[const.NODES] = getItems(c._children)
                    ret.append(cl)
            return ret  
        
        return getItems(response._return._children)
    
    @_wsMethod
    def createCard(self, attributeList, className):
         attributes = []
         for key in attributeList.iterkeys():
            attribute = ns0.attribute_Def('')
            attribute._name = key
            attribute._value = attributeList[key]
            attributes.append(attribute)
         card = ns0.card_Def('', '')
         card._attributeList = attributes
         card._className = className
         card._id = 0
         request = createCard()
         request._cardType = card
         
         response = self.dbfetchSrv.createCard(request)
         return response._return
     
    @_wsMethod 
    def updateCard(self, cardId, attributeList, className):
         attributes = []
         for key in attributeList.iterkeys():
            attribute = ns0.attribute_Def('')
            attribute._name = key
            attribute._value = attributeList[key]
            attributes.append(attribute)
         card = ns0.card_Def('', '')
         card._attributeList = attributes
         card._className = className
         card._id = int(cardId)
         request = updateCard()
         request._card = card
         
         response = self.dbfetchSrv.updateCard(request)
         return response._return
    
    @_wsMethod
    def deleteCard(self, className, cardId):
         request = deleteCard()
         request._className = className
         request._cardId = int(cardId)
         
         response = self.dbfetchSrv.deleteCard(request)
         return response._return

    
    @_wsMethod
    def updateWorkflow(self, attributeList, className):
         attributes = []
         for key in attributeList.iterkeys():
            attribute = ns0.attribute_Def('')
            attribute._name = key
            attribute._value = attributeList[key]
            attributes.append(attribute)
         card = ns0.card_Def('', '')
         card._attributeList = attributes
         card._className = className
         card._id = 0
         request = updateWorkflow()
         request._card = card
         request._completeTask = True
         
         response = self.dbfetchSrv.updateWorkflow(request)         
         return response._return._processid
   
    @_wsMethod
    def uploadAttachment(self,className, objectid, file, filename, category, description):
        request = uploadAttachment()
        request._className = className
        request._objectid = objectid
        request._file = file
        request._filename = filename
        request._category = category
        request._description = description
        response = self.dbfetchSrv.uploadAttachment(request)
        return response._return
        
    @_wsMethod
    def getLookupList(self, lookupType):
         request = getLookupList()
         request._type = lookupType 
         request._parentList = True 
         request._value = ''
         
         response = self.dbfetchSrv.getLookupList(request)
         lookups = []
         lk = dict()
         for i, c  in enumerate(response._return):
            lk[const.DESCRIPTION] = c._description
            lk[const.CODE] = c._code
            lk[const.ID] = c._id
            lk[const.NOTES] = c._notes
            lk[const.PARENT] = c._parent
            lk[const.PARENT_ID] = c._parentId
            lk[const.POSITION] = c._position
            lk[const.TYPE] = c._type
            try:
              lk[const.PARENT_NAME] = c._parent._description
            except:
              lk[const.PARENT_NAME] = ''
            lookups.append(lk)
            lk = dict()
         return lookups
    
    @_wsMethod
    def getAttributeList(self, className, onlyBaseDSP=True):
         request = getAttributeList()
         request._className = className
         
         response = self.dbfetchSrv.getAttributeList(request)
         attributes = []
         attribute = dict()
         for i, c  in enumerate(response._return):
            if onlyBaseDSP:
                if not c._baseDSP:
                    continue
            attribute[const.BASE_DSP] = c._baseDSP
            attribute[const.CLASS_ORDER] = c._classorder
            attribute[const.DEFAULT_VALUE] = c._defaultValue
            attribute[const.DESCRIPTION] = c._description
            attribute[const.FIELD_MODE] = c._fieldmode
            attribute[const.CLASSID] = c._idClass
            attribute[const.ID_DOMAIN] = c._idDomain
            attribute[const.INDEX] = c._index
            attribute[const.INHERITED] = c._inherited
            attribute[const.LENGTH] = c._length
            attribute[const.LOOKUP_TYPE] = c._lookupType
            attribute[const.NAME] = c._name
            attribute[const.NOT_NULL] = c._notnull
            attribute[const.PRECISION] = c._precision
            attribute[const.REFERENCED_CLASSNAME] = c._referencedClassName
            attribute[const.REFERENCED_CLASSID] = c._referencedIdClass
            attribute[const.SCALE] = c._scale
            attribute[const.TYPE] = c._type
            attribute[const.UNIQUE] = c._unique
            attribute[const.VISIBILITY] = c._visibility
            for meta in c._metadata:
                attribute[meta._key] = meta._value
            attributes.append(attribute)
            attribute = dict()
         
         attributes.sort(lambda x, y:cmp(x['index'], y['index']))
         return attributes
     
    @_wsMethod
    def getCMDBuildUser(self, metadataValue):
        refClass = metadataValue.split(".")[0]
        refAttr = metadataValue.split(".")[1]
        
        request = getCardList()
        
        filter1 = ns0.filter_Def(None)
        filter1._name = refAttr
        filter1._operator = 'EQUALS'
        filter1._value = [self.username]
        query1 = ns0.query_Def(None)
        query1._filter = filter1
        
        filter2 = ns0.filter_Def(None)
        filter2._name = "Status"
        filter2._operator = 'EQUALS'
        filter2._value = ['A']     
        query2 = ns0.query_Def(None)
        query2._filter = filter2
        
        query = ns0.query_Def(None)
        filterOp = ns0.filterOperator_Def(None)
        filterOp._operator = 'AND'
        filterOp._subquery = [query1, query2]
        query._filterOperator = filterOp
        
        att1 = ns0.attribute_Def('')     
        att1._name = 'Description'
        
        att2 = ns0.attribute_Def('')
        att2._name = refAttr
        
        attributeList = [att1, att2]
        
        order = ns0.order_Def(None)
        order._columnName = const.ID    
        order._type = 'DESC'
        orderType = [order]
        
        request._className = refClass
        request._limit = 1
        request._offset = 0
        request._queryType = query
        request._orderType = orderType
        request._attributeList = attributeList         
        
        response = self.dbfetchSrv.getCardList(request)
        card = None
        for i, c  in enumerate(response._return._cards):
           card = dict()
           card[const.BEGIN_DATE] = c._beginDate
           card[const.CLASSNAME] = c._className
           card[const.ID] = c._id
           card[const.END_DATE] = c._endDate
           card[const.USER] = c._user
           for a in c._attributeList:
              card[a._name] = a._value
        
        if not card:
            raise Exception(_(u"CMDBuild user required (metadata '%s') but not found: %s") % (metadataValue,self.username))
        
        return card
     
    
    @_wsMethod
    def getCardList(self, name, limit=0, status='A'):
        filter = ns0.filter_Def(None)
        filter._name = "Status"
        filter._operator = 'EQUALS'
        filter._value = [status]
        query = ns0.query_Def(None)
        query._filter = filter
        
        request = getCardList()
        request._className = name
        request._limit = limit
        request._offset = 0
        request._queryType = query                
        
        response = self.dbfetchSrv.getCardList(request)
        cards = []
        card = dict()
        for i, c  in enumerate(response._return._cards):
            card[const.BEGIN_DATE] = c._beginDate
            card[const.CLASSNAME] = c._className
            card[const.ID] = c._id
            card[const.END_DATE] = c._endDate
            card[const.USER] = c._user
            for a in c._attributeList:
               card[a._name] = a._value
            cards.append(card)
            card = dict()
        # cards.sort(lambda x,y:cmp(x['Description'],y['Description']))
        return cards
        
    
    @_wsMethod
    def getCardListQuery(self, name, attributes, offset, limit, flowstatus):
         request = getCardList()
         subquery = []
         if flowstatus != None and flowstatus != '' and int(flowstatus) > 0:
            filter = ns0.filter_Def(None)
            filter._name = "FlowStatus"
            filter._operator = 'EQUALS'
            filter._value = [flowstatus]
            query = ns0.query_Def(None)
            query._filter = filter
            subquery.append(query)
         else:
            filter = ns0.filter_Def(None)
            filter._name = "Status"
            filter._operator = 'EQUALS'
            filter._value = ['A']
            query = ns0.query_Def(None)
            query._filter = filter
            subquery.append(query)
        
         query = ns0.query_Def(None)
         filterOp = ns0.filterOperator_Def(None)
         filterOp._operator = 'AND'
         filterOp._subquery = subquery
         query._filterOperator = filterOp
        
         attributeList = []
         for attribute in attributes:
             att = ns0.attribute_Def('')
             att._name = attribute[const.NAME]
             attributeList.append(att)
        
         order = ns0.order_Def(None)
         order._columnName = const.ID
         order._type = 'DESC'
         orderType = [order]
        
         request._className = name
         request._limit = int(limit)
         request._offset = int(offset)
         request._queryType = query
         request._orderType = orderType
         request._attributeList = attributeList
         
         response = self.dbfetchSrv.getCardList(request)
         cards = []
         card = dict()
         attributi = dict()
         for i, c  in enumerate(response._return._cards):
            card[const.BEGIN_DATE] = c._beginDate
            card[const.END_DATE] = c._endDate
            card[const.CLASSNAME] = c._className
            card[const.ID] = c._id
            card[const.USER] = c._user
            for a in c._attributeList:
               card[a._name] = a._value
            cards.append(card)
            card = dict()
         
         totalRows = response._return._totalRows
         response = [totalRows, cards]
         return response

     
    @_wsMethod
    def getReference(self, className):
        request = getReference()
        
        # class
        request._className = className
        
        # order
        order = ns0.order_Def(None)
        order._columnName = const.DESCRIPTION
        order._type = 'ASC'
        request._orderType = [order]

        response = self.dbfetchSrv.getReference(request)
        cards = []
        card = dict()
        for i, c  in enumerate(response._return):
            card[const.DESCRIPTION] = c._description
            card[const.ID] = c._id
            cards.append(card)
            card = dict()
        # cards.sort(lambda x,y:cmp(x['Description'],y['Description']))
        return cards
    
    
    @_wsMethod
    def searchCards(self, className, attributes, offset, limit, status):
         searchDict = self.session.get(const.SESSION_SEARCHDIC)
         if searchDict == None:
             raise Exception("Search parameters not found in session")
          
         request = getCardList()
         subquery = []
         for key in searchDict.iterkeys():
            query = ns0.query_Def(None)
            filter = ns0.filter_Def(None)
            filter._name = key
            entry = searchDict[key].split('|')
            filter._operator = entry[0]
            if entry[1] != '':
             filter._value = [entry[1]]
             query._filter = filter
             subquery.append(query) 
        
         # added to get all results if search fields are empty
         if status != '' and int(status) > 0:
            filter = ns0.filter_Def(None)
            filter._name = "FlowStatus"
            filter._operator = 'EQUALS'
            filter._value = [status]
            query = ns0.query_Def(None)
            query._filter = filter
            subquery.append(query)
         else:     
             query = ns0.query_Def(None)
             filter1 = ns0.filter_Def(None)
             filter1._name = 'Status'
             filter1._value = ['A']
             filter1._operator = 'EQUALS'
             query._filter = filter1
             subquery.append(query)
         
         query = ns0.query_Def(None)
         filterOp = ns0.filterOperator_Def(None)
         filterOp._operator = 'AND'
         filterOp._subquery = subquery
         query._filterOperator = filterOp
        
         attributeList = []
         for l in attributes:
          att = ns0.attribute_Def('')
          att._name = l['name']
          attributeList.append(att)
        
         order = ns0.order_Def(None)
         order._columnName = const.ID
         order._type = 'DESC'
         orderType = [order]
        
         request._className = className
         request._limit = limit
         request._offset = offset
         request._queryType = query
         request._orderType = orderType
         request._attributeList = attributeList
         
         response = self.dbfetchSrv.getCardList(request)
         cards = []
         card = dict()
         for i, c  in enumerate(response._return._cards):
            card[const.BEGIN_DATE] = c._beginDate
            card[const.END_DATE] = c._endDate
            card[const.CLASSNAME] = c._className
            card[const.ID] = c._id
            card[const.TOTAL_ROWS] = response._return._totalRows
            card[const.USER] = c._user
            for a in c._attributeList:
               card[a._name] = a._value
            cards.append(card)
            card = dict()
         
         totalRows = response._return._totalRows
         response = [totalRows, cards]
         return response 
    
    @_wsMethod
    def getCard(self, id, className):
         request = getCard()
         request._cardId = id
         request._className = className
         response = self.dbfetchSrv.getCard(request)
         cards = []
         card = dict()
         c = response._return 
         card[const.BEGIN_DATE] = c._beginDate
         card[const.END_DATE] = c._endDate
         card[const.CLASSNAME] = c._className
         card[const.ID] = c._id
         card[const.USER] = c._user
         for attribute in c._attributeList:
              card[attribute._name] = attribute._value
         return card
    
    @_wsMethod
    def getActivityObjects(self, className):
         request = getActivityObjects()
         request._className = className
         response = self.dbfetchSrv.getActivityObjects(request)
         attributes = []
         attribute = dict()         
         for i, c  in enumerate(response._return._attributes):
            attribute[const.BASE_DSP] = c._baseDSP
            attribute[const.CLASS_ORDER] = c._classorder
            attribute[const.DEFAULT_VALUE] = c._defaultValue
            attribute[const.DESCRIPTION] = c._description
            attribute[const.FIELD_MODE] = c._fieldmode
            attribute[const.CLASSID] = c._idClass
            attribute[const.ID_DOMAIN] = c._idDomain
            attribute[const.INDEX] = c._index
            attribute[const.INHERITED] = c._inherited
            attribute[const.LENGTH] = c._length
            attribute[const.LOOKUP_TYPE] = c._lookupType
            attribute[const.NAME] = c._name
            attribute[const.NOT_NULL] = c._notnull
            attribute[const.PRECISION] = c._precision
            attribute[const.REFERENCED_CLASSNAME] = c._referencedClassName
            attribute[const.REFERENCED_CLASSID] = c._referencedIdClass
            attribute[const.SCALE] = c._scale
            attribute[const.TYPE] = c._type
            attribute[const.UNIQUE] = c._unique
            attribute[const.VISIBILITY] = c._visibility
            for meta in c._metadata:
                attribute[meta._key] = meta._value
            attributes.append(attribute)
            attribute = dict()
         return attributes 
     
